# Evaluation script for CUDA kernel
# 63_conv_standard_2D__square_input__square_kernel
# Evaluation script for CUDA kernel
# 12_Matmul_with_diagonal_matrices_
import os
import torch
import argparse
from torch.utils.cpp_extension import load
from torch.utils._pytree import tree_map
import importlib.util
from torch.utils.benchmark import Timer


def easy_to_device(pytree, device):
    return tree_map(
        lambda x: x.to(device) if isinstance(x, torch.Tensor) else x, pytree
    )


def load_module_from_path(path):
    spec = importlib.util.spec_from_file_location("module", path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--op_atol", type=float, default=1e-3)
    parser.add_argument("--op_rtol", type=float, default=1e-1)
    parser.add_argument("--rep_time", type=int, default=10000)
    parser.add_argument("--warmup_time", type=int, default=25)
    args = parser.parse_args()

    # Get task name from info.txt
    with open("task/info.txt", "r") as f:
        task_name = f.readline().strip()
        task_name = "_".join(task_name.split("_")[1:])  # Remove problem ID

    # Import the task module
    task_files = [f for f in os.listdir("task") if f.endswith("_functional.py")]
    if not task_files:
        raise RuntimeError("No functional task file found")

    task = load_module_from_path(os.path.join("task", task_files[0]))

    # Initialize model and inputs
    device_1 = torch.device("cuda:0")
    device_2 = torch.device("cuda:1")
    torch.manual_seed(0)
    inputs = task.get_inputs()
    init_inputs = task.get_init_inputs()
    model = task.Model(*init_inputs)

    # Load CUDA kernel
    kernel_files = [f for f in os.listdir("kernel") if f.endswith(".cu")]
    if not kernel_files:
        raise RuntimeError("No CUDA kernel file found")

    cuda_module = load(
        name=task_name,
        sources=[os.path.join("kernel", kernel_files[0])],
        extra_cuda_cflags=["-O3", "--use_fast_math"],
        with_cuda=True,
        verbose=True,
    )

    # Test for correctness
    with torch.no_grad():
        cuda_output = model.to(device_1)(
            *easy_to_device(inputs, device_1), fn=cuda_module.forward
        )
        torch_output = model.to(device_2)(
            *easy_to_device(inputs, device_2), fn=task.module_fn
        )

    correct = torch.allclose(
        torch_output.cpu(),
        cuda_output.cpu(),
        rtol=args.op_rtol,
        atol=args.op_atol,
    )
    max_diff = torch.max(torch.abs(torch_output.cpu() - cuda_output.cpu())).item()
    print(f"Tested CUDA kernel - Correct: {correct}, Max Diff: {max_diff}")

    if correct:
        # Evaluate CUDA kernel performance
        cuda_timer = Timer(
            stmt="model(*inputs, fn=cuda_module.forward)",
            globals={
                "model": model.to(device_1),
                "inputs": easy_to_device(inputs, device_1),
                "cuda_module": cuda_module,
            },
        )
        cuda_runtime = cuda_timer.timeit(args.rep_time).mean * 1000
        print(f"Evaluated CUDA kernel - Runtime: {cuda_runtime:.3f} ms")

        # Evaluate PyTorch baseline performance
        torch_timer = Timer(
            stmt="model(*inputs, fn=task.module_fn)",
            globals={
                "model": model.to(device_2),
                "inputs": easy_to_device(inputs, device_2),
                "task": task,
            },
        )
        torch_runtime = torch_timer.timeit(args.rep_time).mean * 1000
        print(f"Evaluated PyTorch baseline - Runtime: {torch_runtime:.3f} ms")

        # Evaluate torch compile performance
        torch_fn = task.module_fn
        compile_fn = torch.compile(torch_fn, mode="max-autotune")
        torch_compile_timer = Timer(
            stmt="model(*inputs, fn=compile_fn)",
            globals={
                "model": model.to(device_2),
                "inputs": easy_to_device(inputs, device_2),
                "compile_fn": compile_fn,
            },
        )

        torch_compile_runtime = torch_compile_timer.timeit(args.rep_time).mean * 1000
        print(f"Evaluated torch compile - Runtime: {torch_compile_runtime:.3f} ms")

        print(f"Speedup over PyTorch: {torch_runtime/cuda_runtime:.2f}x")
        print(f"Speedup over torch compile: {torch_compile_runtime/cuda_runtime:.2f}x")

        import json

        # Store the speedup times as a json file
        file_path = os.path.join(os.path.dirname(__file__), "speedup_times.json")
        with open(file_path, "w") as f:
            json.dump(
                {
                    "max_diff": max_diff,
                    "cuda_runtime": cuda_runtime,
                    "torch_runtime": torch_runtime,
                    "torch_compile_runtime": torch_compile_runtime,
                    "speedup_over_pytorch": torch_runtime / cuda_runtime,
                    "speedup_over_torch_compile": torch_compile_runtime / cuda_runtime,
                },
                f,
            )
        print(f"Speedup times stored in {file_path}")


if __name__ == "__main__":
    main()
